@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2024 Michal Trojnara <Michal.Trojnara@stunnel.org>
+ *   Copyright (C) 1998-2025 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -430,7 +430,7 @@ int ocsp_stapling(SERVICE_OPTIONS *opt) {
         /* cache the newly fetched OCSP response */
         CRYPTO_THREAD_write_lock(opt->ocsp_response_lock);
         opt->ocsp_response_len=response_len;
-        opt->ocsp_response_der=str_alloc((size_t)response_len);
+        opt->ocsp_response_der=str_alloc_detached((size_t)response_len);
         memcpy(opt->ocsp_response_der, response_der, (size_t)response_len);
         CRYPTO_THREAD_unlock(opt->ocsp_response_lock);
         s_log(LOG_DEBUG, "OCSP: Response cached");
@@ -648,6 +648,7 @@ NOEXPORT int ocsp_get_response(SERVICE_OPTIONS *opt, OCSP_PARAMS *params) {
      * - rewrite the network.c interface to use this structure */
     c=str_alloc(sizeof(CLI));
     c->fds=s_poll_alloc();
+    c->fd=INVALID_SOCKET;
     c->exception_pointer=&exception_buffer;
     if(setjmp(exception_buffer)) {
         s_poll_free(c->fds);
